@@ -5,7 +5,7 @@ package OpenAPI::Modern;
 # ABSTRACT: Validate HTTP requests and responses against an OpenAPI document
 # KEYWORDS: validation evaluation JSON Schema OpenAPI Swagger HTTP request response
 
-our $VERSION = '0.003';
+our $VERSION = '0.004';
 
 use 5.020;  # for fc, unicode_strings features
 use Moo;
@@ -17,10 +17,12 @@ no if "$]" >= 5.033001, feature => 'multidimensional';
 no if "$]" >= 5.033006, feature => 'bareword_filehandles';
 use Carp 'croak';
 use Safe::Isa;
-use Ref::Util 'is_plain_hashref';
+use Ref::Util qw(is_plain_hashref is_plain_arrayref);
+use List::Util 'first';
+use Scalar::Util 'looks_like_number';
 use Feature::Compat::Try;
 use JSON::Schema::Modern 0.527;
-use JSON::Schema::Modern::Utilities qw(jsonp canonical_uri E abort);
+use JSON::Schema::Modern::Utilities qw(jsonp canonical_uri E abort get_type);
 use JSON::Schema::Modern::Document::OpenAPI;
 use MooX::HandlesVia;
 use MooX::TypeTiny 0.002002;
@@ -88,75 +90,50 @@ sub validate_request ($self, $request, $options) {
   my $captures = $options->{path_captures};
   croak 'missing parameter path_template' if not length $path_template;
   croak 'missing parameter captures ' if not is_plain_hashref($captures);
-  my $method = lc $request->method;
 
-  my $schema = $self->openapi_document->schema;
-  my $path_item = $schema->{paths}{$path_template};
-
-  $state->{schema_path} = jsonp('/paths', $path_template);
-  if (not $path_item) {
-    my $valid = E($state, 'missing path-item "%s"', $path_template);
-    return $self->_result($state);
-  }
-
-  my $operation = $path_item->{ $method };
-  if (not $operation) {
-      my $valid = E({ %$state, _schema_path_suffix => $method }, 'missing operation');
-      return $self->_result($state);
-  }
-
-  # PARAMETERS
   try {
-    # { $in => { $name => 1 } }  as we process each one.
+    my $method = lc $request->method;
+    my $schema = $self->openapi_document->schema;
+
+    my $path_item = $schema->{paths}{$path_template};
+    abort({ %$state, keyword => 'paths', _schema_path_suffix => $path_template },
+      'missing path-item "%s"', $path_template) if not $path_item;
+
+    $state->{schema_path} = jsonp('/paths', $path_template);
+    my $operation = $path_item->{$method};
+    abort({ %$state, keyword => $method }, 'missing entry for HTTP method "%s"', $method)
+      if not $operation;
+
+    # PARAMETERS
+    # { $in => { $name => 'path-item'|$method } }  as we process each one.
     my $request_parameters_processed;
 
     # first, consider parameters at the operation level.
-
-    #TODO: we can wrap another loop around these two sections because they are so similar.
-
-    foreach my $idx (0 .. ($operation->{parameters}//[])->$#*) {
-      my $state = {
-        %$state,
-        schema_path => jsonp($state->{schema_path}, $method, 'parameters', $idx),
-      };
-      my $param_obj = $operation->{parameters}[$idx];
-      while (my $ref = $param_obj->{'$ref'}) {
-        $param_obj = $self->_resolve_ref($ref, $state);
-      }
-
-      ++$request_parameters_processed->{$param_obj->{in}}{
-        $param_obj->{in} eq 'header' ? lc($param_obj->{name}) : $param_obj->{name}
-      };
-      $state->{data_path} = jsonp($state->{data_path}, $param_obj->{in}, $param_obj->{name});
-      my $valid =
-          $param_obj->{in} eq 'path' ? $self->_validate_path_parameter($state, $param_obj, $captures)
-        : $param_obj->{in} eq 'query' ? $self->_validate_query_parameter($state, $param_obj, $request->uri)
-        : $param_obj->{in} eq 'header' ? $self->_validate_header_parameter($state, $param_obj, $request->headers)
-        : $param_obj->{in} eq 'cookie' ? $self->_validate_cookie_parameter($state, $param_obj, $request)
-        : abort($state, 'unrecognized "in" value "%s"', $param_obj->{in});
-    }
-
     # parameters at the path-item level are also considered, if not already seen at the operation level
+    foreach my $section ($method, 'path-item') {
+      foreach my $idx (0 .. (($section eq $method ? $operation : $path_item)->{parameters}//[])->$#*) {
+        my $state = { %$state, schema_path => jsonp($state->{schema_path},
+          ($section eq $method ? $method : ()), 'parameters', $idx) };
+        my $param_obj = ($section eq $method ? $operation : $path_item)->{parameters}[$idx];
+        while (my $ref = $param_obj->{'$ref'}) {
+          $param_obj = $self->_resolve_ref($ref, $state);
+        }
 
-    foreach my $idx (0 .. ($path_item->{parameters}//[])->$#*) {
-      my $state = {
-        %$state,
-        schema_path => jsonp('/paths', $path_template, 'parameters', $idx),
-      };
-      my $param_obj = $path_item->{parameters}[$idx];
-      while (my $ref = $param_obj->{'$ref'}) {
-        $param_obj = $self->_resolve_ref($ref, $state);
+        my $fc_name = $param_obj->{in} eq 'header' ? fc($param_obj->{name}) : $param_obj->{name};
+
+        abort($state, 'duplicate path parameter "%s"', $param_obj->{name})
+          if ($request_parameters_processed->{$param_obj->{in}}{$fc_name} // '') eq $section;
+        next if exists $request_parameters_processed->{$param_obj->{in}}{$fc_name};
+        $request_parameters_processed->{$param_obj->{in}}{$fc_name} = $section;
+
+        $state->{data_path} = jsonp($state->{data_path}, $param_obj->{in}, $param_obj->{name});
+        my $valid =
+            $param_obj->{in} eq 'path' ? $self->_validate_path_parameter($state, $param_obj, $captures)
+          : $param_obj->{in} eq 'query' ? $self->_validate_query_parameter($state, $param_obj, $request->uri)
+          : $param_obj->{in} eq 'header' ? $self->_validate_header_parameter($state, $param_obj->{name}, $param_obj, $request->headers)
+          : $param_obj->{in} eq 'cookie' ? $self->_validate_cookie_parameter($state, $param_obj, $request)
+          : abort($state, 'unrecognized "in" value "%s"', $param_obj->{in});
       }
-      next if $request_parameters_processed->{$param_obj->{in}}{
-        $param_obj->{in} eq 'header' ? lc($param_obj->{name}) : $param_obj->{name}
-      }++;
-      $state->{data_path} = jsonp($state->{data_path}, $param_obj->{in}, $param_obj->{name});
-      my $valid =
-          $param_obj->{in} eq 'path' ? $self->_validate_path_parameter($state, $param_obj, $captures)
-        : $param_obj->{in} eq 'query' ? $self->_validate_query_parameter($state, $param_obj, $request->uri)
-        : $param_obj->{in} eq 'header' ? $self->_validate_header_parameter($state, $param_obj, $request->headers)
-        : $param_obj->{in} eq 'cookie' ? $self->_validate_cookie_parameter($state, $param_obj, $request)
-        : abort($state, 'unrecognized "in" value "%s"', $param_obj->{in});
     }
 
     if (my $body_obj = $operation->{requestBody}) {
@@ -167,7 +144,12 @@ sub validate_request ($self, $request, $options) {
         $body_obj = $self->_resolve_ref($ref, $state);
       }
 
-      my $valid = $self->_validate_body($state, $body_obj, $request);
+      if ($body_obj->{required} and not length($request->content_ref->$*)) {
+        ()= E({ %$state, keyword => 'required' }, 'request body is required but missing');
+      }
+      else {
+        ()= $self->_validate_body_content($state, $body_obj->{content}, $request);
+      }
     }
   }
   catch ($e) {
@@ -185,11 +167,80 @@ sub validate_request ($self, $request, $options) {
   return $self->_result($state);
 }
 
-sub validate_response ($self, $response, @) {
-  return JSON::Schema::Modern::Result->new(
-    output_format => $self->evaluator->output_format,
-    valid => true,
-  );
+# at the moment, we rely on these values being provided in $options:
+# - path_template
+sub validate_response ($self, $response, $options) {
+  my $state = {
+    data_path => '/response',
+    initial_schema_uri => $self->openapi_uri,   # the canonical URI as of the start or last $id, or the last traversed $ref
+    traversed_schema_path => '',    # the accumulated traversal path as of the start, or last $id, or up to the last traversed $ref
+    schema_path => '',              # the rest of the path, since the last $id or the last traversed $ref
+    errors => [],
+  };
+
+  my $path_template = $options->{path_template};
+  croak 'missing parameter path_template' if not length $path_template;
+
+  try {
+    my $schema = $self->openapi_document->schema;
+    abort({ %$state, keyword => 'paths', _schema_path_suffix => $path_template },
+        'missing path-item "%s"', $path_template)
+      if not $schema->{paths}{$path_template};
+
+    $state->{schema_path} = jsonp('/paths', $path_template);
+    my $method = lc $response->request->method;
+    my $operation = $schema->{paths}{$path_template}{$method};
+    abort({ %$state, _schema_path_suffix => $method }, 'missing operation')
+      if not $operation;
+
+    return $self->_result($state) if not exists $operation->{responses};
+
+    $state->{schema_path} = jsonp($state->{schema_path}, $method);
+
+    my $response_property = first { exists $operation->{responses}{$_} }
+      $response->code, substr(sprintf('%03s', $response->code), 0, -2).'XX', 'default';
+
+    if (not $response_property) {
+      ()= E({ %$state, keyword => 'responses' }, 'no response object found for code %s', $response->code);
+      return $self->_result($state);
+    }
+
+    my $response_obj = $operation->{responses}{$response_property};
+    $state->{schema_path} = jsonp($state->{schema_path}, 'responses', $response_property);
+    while (my $ref = $response_obj->{'$ref'}) {
+      $response_obj = $self->_resolve_ref($ref, $state);
+    }
+
+    foreach my $header_name (keys(($response_obj->{headers}//{})->%*)) {
+      next if fc $header_name eq fc 'Content-Type';
+      my $state = { %$state, schema_path => jsonp($state->{schema_path}, 'headers', $header_name) };
+      my $header_obj = $response_obj->{headers}{$header_name};
+      while (my $ref = $header_obj->{'$ref'}) {
+        $header_obj = $self->_resolve_ref($ref, $state);
+      }
+
+      ()= $self->_validate_header_parameter({ %$state,
+          data_path => jsonp($state->{data_path}, 'header', $header_name) },
+        $header_name, $header_obj, $response->headers);
+    }
+
+    ()= $self->_validate_body_content({ %$state, data_path => jsonp($state->{data_path}, 'body') },
+        $response_obj->{content}, $response)
+      if exists $response_obj->{content};
+  }
+  catch ($e) {
+    if ($e->$_isa('JSON::Schema::Modern::Result')) {
+      return $e;
+    }
+    elsif ($e->$_isa('JSON::Schema::Modern::Error')) {
+      push @{$state->{errors}}, $e;
+    }
+    else {
+      E($state, 'EXCEPTION: '.$e);
+    }
+  }
+
+  return $self->_result($state);
 }
 
 ######## NO PUBLIC INTERFACES FOLLOW THIS POINT ########
@@ -199,20 +250,12 @@ sub _validate_path_parameter ($self, $state, $param_obj, $captures) {
   return E({ %$state, keyword => 'content' }, 'content not yet supported')
     if exists $param_obj->{content};
 
-  # 'required' must be true for path parameters
+  # 'required' is always true for path parameters
   return E({ %$state, keyword => 'required' }, 'missing path parameter: %s', $param_obj->{name})
     if not exists $captures->{$param_obj->{name}};
 
   $state = { %$state, schema_path => jsonp($state->{schema_path}, 'schema') };
-
-  my $result = $self->evaluator->evaluate(
-    $captures->{$param_obj->{name}},
-    canonical_uri($state),
-    { data_path => $state->{data_path}, traversed_schema_path => $state->{schema_path} },
-  );
-  push $state->{errors}->@*, $result->errors;
-  push $state->{annotations}->@*, $result->annotations if $self->evaluator->collect_annotations;
-  return !!$result;
+  $self->_evaluate_subschema($captures->{$param_obj->{name}}, $param_obj->{schema}, $state);
 }
 
 sub _validate_query_parameter ($self, $state, $param_obj, $uri) {
@@ -236,60 +279,51 @@ sub _validate_query_parameter ($self, $state, $param_obj, $uri) {
   # TODO: check 'allowReserved': it cannot be supported if we use proper URL encoding
 
   $state = { %$state, schema_path => jsonp($state->{schema_path}, 'schema') };
-  my $result = $self->evaluator->evaluate(
-    $query_params->{$param_obj->{name}},
-    canonical_uri($state),
-    { data_path => $state->{data_path}, traversed_schema_path => $state->{schema_path} },
-  );
-  push $state->{errors}->@*, $result->errors;
-  push $state->{annotations}->@*, $result->annotations if $self->evaluator->collect_annotations;
-  return !!$result;
+  $self->_evaluate_subschema($query_params->{$param_obj->{name}}, $param_obj->{schema}, $state);
 }
 
 # validates a header, from either the request or the response
-sub _validate_header_parameter ($self, $state, $param_obj, $headers) {
+sub _validate_header_parameter ($self, $state, $header_name, $header_obj, $headers) {
   return E({ %$state, keyword => 'content' }, 'content not yet supported')
-      if exists $param_obj->{content};
+    if exists $header_obj->{content};
 
-  # NOTE: for now, we will only support a single value, as a string.
-  my @values = $headers->header($param_obj->{name});
+  # NOTE: for now, we will only support a single value.
+  my @values = $headers->header($header_name);
   if (not @values) {
-    return E({ %$state, keyword => 'required' }, 'missing header: %s', $param_obj->{name})
-      if $param_obj->{required};
+    return E({ %$state, keyword => 'required' }, 'missing header: %s', $header_name)
+      if $header_obj->{required};
     return 1;
   }
 
   $state = { %$state, schema_path => jsonp($state->{schema_path}, 'schema') };
-  my $result = $self->evaluator->evaluate(
-    $values[0],
-    canonical_uri($state),
-    { data_path => $state->{data_path}, traversed_schema_path => $state->{schema_path} },
-  );
-  push $state->{errors}->@*, $result->errors;
-  push $state->{annotations}->@*, $result->annotations if $self->evaluator->collect_annotations;
-  return !!$result;
+  $self->_evaluate_subschema($values[0], $header_obj->{schema}, $state);
 }
 
 sub _validate_cookie_parameter ($self, $state, $param_obj, $request) {
   return E($state, 'cookie parameters not yet supported');
 }
 
-sub _validate_body ($self, $state, $body_obj, $message) {
-  my $content_ref = $message->content_ref;
-  my $type = $message->isa('HTTP::Request') ? 'request'
-    : $message->isa('HTTP::Response') ? 'response'
-    : return E($state, 'unrecognized object type %s', ref($message));
-
-  return E({ %$state, keyword => 'required' }, '%s body is required but missing', $type)
-    if $body_obj->{required} and not length($content_ref->$*);
-
+sub _validate_body_content ($self, $state, $content_obj, $message) {
   my $content_type = $message->content_type;
 
   # TODO: respect wildcard entries in the openapi document
   return E({ %$state, keyword => 'content' }, 'incorrect Content-Type "%s"', $content_type)
-    if not exists $body_obj->{content}{$content_type};
+    if not exists $content_obj->{$content_type};
 
-  die 'found encoding: TODO' if exists $body_obj->{content}{$content_type}{encoding};
+  if (exists $content_obj->{$content_type}{encoding}) {
+    my $state = { %$state, schema_path => jsonp($state->{schema_path}, 'content', $content_type) };
+    # "The key, being the property name, MUST exist in the schema as a property."
+    foreach my $property (keys $content_obj->{$content_type}{encoding}->%*) {
+      ()= E({ $state, schema_path => jsonp($state->{schema_path}, 'schema', 'properties', $property) },
+          'encoding property "%s" requires a matching property definition in the schema')
+        if not exists(($content_obj->{$content_type}{schema}{properties}//{})->{$property});
+    }
+
+    # "The encoding object SHALL only apply to requestBody objects when the media type is multipart or
+    # application/x-www-form-urlencoded."
+    return E({ %$state, keyword => 'encoding' }, 'encoding not yet supported')
+      if $content_type =~ m{^multipart/} or $content_type eq 'application/x-www-form-urlencoded';
+  }
 
   # undoes the Content-Encoding header
   my $decoded_content_ref = $message->decoded_content(ref => 1);
@@ -306,14 +340,7 @@ sub _validate_body ($self, $state, $body_obj, $message) {
   $decoded_content_ref = $media_type_decoder->($decoded_content_ref);
 
   $state = { %$state, schema_path => jsonp($state->{schema_path}, 'content', $content_type, 'schema') };
-  my $result = $self->evaluator->evaluate(
-    $decoded_content_ref->$*,
-    canonical_uri($state),
-    { data_path => $state->{data_path}, traversed_schema_path => $state->{schema_path} },
-  );
-  push $state->{errors}->@*, $result->errors;
-  push $state->{annotations}->@*, $result->annotations if $self->evaluator->collect_annotations;
-  return !!$result;
+  $self->_evaluate_subschema($decoded_content_ref->$*, $content_obj->{$content_type}{schema}, $state);
 }
 
 # wrap a result object around the errors
@@ -342,6 +369,38 @@ sub _resolve_ref ($self, $ref, $state) {
   return $schema_info->{schema};
 }
 
+# evaluates data against the subschema at the current state location
+sub _evaluate_subschema ($self, $data, $schema, $state) {
+  return 1 if is_plain_hashref($schema) ? !keys(%$schema) : $schema; # true schema
+
+  # treat numeric-looking data as a string, unless "type" explicitly requests number or integer.
+  my $type = get_type($data);
+  if ((grep $type eq $_, qw(string number)) and looks_like_number($data)) {
+    if (is_plain_hashref($schema) and exists $schema->{type} and not is_plain_arrayref($schema->{type})
+        and grep $schema->{type} eq $_, qw(number integer)) {
+      $data = $data+0;
+    }
+    else {
+      $data = $data.'';
+    }
+  }
+
+  # TODO: also handle multi-valued elements like headers and query parameters, when type=array requested
+  # (and possibly coerce their numeric-looking elements as well)
+
+  my $result = $self->evaluator->evaluate(
+    $data, canonical_uri($state),
+    {
+      data_path => $state->{data_path},
+      traversed_schema_path => $state->{traversed_schema_path}.$state->{schema_path},
+    },
+  );
+
+  push $state->{errors}->@*, $result->errors;
+  push $state->{annotations}->@*, $result->annotations if $self->evaluator->collect_annotations;
+  return !!$result;
+}
+
 1;
 
 __END__
@@ -356,7 +415,7 @@ OpenAPI::Modern - Validate HTTP requests and responses against an OpenAPI docume
 
 =head1 VERSION
 
-version 0.003
+version 0.004
 
 =head1 SYNOPSIS
 
@@ -391,7 +450,7 @@ version 0.003
                 properties:
                   hello:
                     type: string
-                    pattern: ^[0-9]+\$
+                    pattern: ^[0-9]+$
         responses:
           200:
             description: success
@@ -447,7 +506,7 @@ than features that seem to work but actually cut corners for simplicity.
 
 =for Pod::Coverage BUILDARGS
 
-=for :stopwords schemas jsonSchemaDialect metaschema subschema
+=for :stopwords schemas jsonSchemaDialect metaschema subschema perlish
 
 =head1 CONSTRUCTOR ARGUMENTS
 
@@ -499,7 +558,7 @@ The L<JSON::Schema::Modern> object to use for all URI resolution and JSON Schema
     $request,
     {
       path_captures => { arg1 => 1, arg2 => 2 },
-      path_spec => '/foo/{arg1}/bar/{arg2}',
+      path_template => '/foo/{arg1}/bar/{arg2}',
     },
   );
 
@@ -525,11 +584,19 @@ C<path_captures>: a hashref mapping placeholders in the path to their actual val
   $result = $openapi->validate_response(
     $response,
     {
-      path_spec => '/foo/{arg1}/bar/{arg2}',
+      path_template => '/foo/{arg1}/bar/{arg2}',
     },
   );
 
-Not yet implemented.
+The second argument is a hashref that contains extra information about the request. Possible values include:
+
+=over 4
+
+=item *
+
+C<path_template>: a string representing the request URI, with placeholders in braces (e.g. C</pets/{petId}>); see L<https://spec.openapis.org/oas/v3.1.0#paths-object>.
+
+=back
 
 =head1 ON THE USE OF JSON SCHEMAS
 
@@ -547,6 +614,15 @@ as such, json pointers are relative to the B<root> of the document, not the root
 itself. References to other documents are also permitted, provided those documents have been loaded
 into the evaluator in advance (see L<JSON::Schema::Modern/add_schema>).
 
+Values are generally treated as strings for the purpose of schema evaluation. However, if the top
+level of the schema contains C<"type": "number"> or C<"type": "integer">, then the value will be
+(attempted to be) coerced into a number before being passed to the JSON Schema evaluator.
+Type coercion will B<not> be done if the C<type> keyword is omitted.
+This lets you use numeric keywords such as C<maximum> and C<multipleOf> in your schemas.
+It also resolves inconsistencies that can arise when request and response objects are created
+manually in a test environment (as opposed to being parsed from incoming network traffic) and can
+therefore inadvertently contain perlish numbers rather than strings.
+
 =head1 LIMITATIONS
 
 Only certain permutations of OpenAPI specifications and are supported at this time:
@@ -581,10 +657,6 @@ for query and header parameters, only the first value of each name is considered
 
 media-type encodings in parameters are not yet supported
 
-=item *
-
-responses aren't checked at all yet
-
 =back
 
 =head1 SEE ALSO
@@ -623,6 +695,9 @@ Bugs may be submitted through L<https://github.com/karenetheridge/JSON-Schema-Mo
 
 I am also usually active on irc, as 'ether' at C<irc.perl.org> and C<irc.libera.chat>.
 
+You can also find me on the L<JSON Schema Slack server|json-schema.slack.com> and L<OpenAPI Slack
+server|open-api.slack.com>, which are also great resources for finding help.
+
 =head1 AUTHOR
 
 Karen Etheridge <ether@cpan.org>
